#ifndef Image_h
#define Image_h

//----------------------------------------------------------------
// Class Image
//
// Interface definition for Image class
//----------------------------------------------------------------


#include <stdio.h>
#include <math.h>
#include <string.h>
#include "Dal.h"
//#include "rfftw.h"
#include "fftw3.h"
#include "Error.h"
#include "errstr.h"
#include "commondef.h"



#define MIN_DIV          1.0E-12
#define MAX_SIZE         4096
#define MAX_MEDIAN       25
#define BLANK            0.0

#define FLEN_KEYWORD    9  /* max length of a keyword */
#define FLEN_CARD      81  /* length of a FITS header card */
#define FLEN_VALUE     71  /* max length of a keyword value string */
#define FLEN_COMMENT   73  /* max length of a keyword comment string */
#define FLEN_ERRMSG    81  /* max length of a FITSIO error message */ 
#define FLEN_STATUS    31  /* max length of a FITSIO status text string */

//----------------------------------------------------------//
// These varaiables are obsolete in FTTW  3.x,  
// but are still used in the ewavelet code to keep track of 
// the transformation direction (GL, 09/04/2009):
#define FFTW_REAL_TO_COMPLEX -1
#define FFTW_COMPLEX_TO_REAL  1
//-------------------------------------------------//

typedef enum projection{
  UNKNOWN = 0, RADECTAN = 1, RADECSIN = 2, RADECSTG = 3 , RADECARC=4
} Projection;


typedef enum coordinate_type{
  RAW_XY = 0, DET_XY = 1, SKY_XY = 2, RA_DEC=3, L_B = 4
} Coordinate_type;


class Image{

 public:

  Image( void );
  Image( const Image &im ); // copy constructor
  Image( int xsize, int ysize, float val=0.0 );
  ~Image();
  Image &operator=(Image const &im );

  Status operator+=( const Image &im );
  Status operator-=( const Image &im );
  Status operator*=( const Image &im );
  Status operator/=( const Image &im );
  Status operator*=( double a );
  Status operator/=( double a );
  Status operator+=( double a );

  void reset_keywords();
  Status copy_keywords( Image const &obs);

  Status readImage(  const string &setname, unsigned int blocknr=0 );
  Status writeImage( const string &setname) const;
  void printInfo(void) const;

  double total( void ) const;
  double total( unsigned int i1, unsigned int j1, 
		unsigned int i2, unsigned int j2 ) const;
  float maximum( void ) const;
  float minimum( void ) const;
  float mean( void ) const;
  Status cage(float minval, float maxval);
  Status clip(float minval, float maxval);
  float  get( unsigned int i, unsigned int j ) const;
  float getExposure( void ) const;
  Status setpixel( unsigned int i, unsigned int j, float val );
  void getnxny( unsigned int *m, unsigned int *n ) const;

  Status windowfilter(const Image &im, float *window, 
		      int n, bool renormalize,
		      int i1=-1, int j1=-1, int i2=-1, int j2=-1);
  Status gaussianfilter(const Image &im, float sigma, 
			int i1=-1, int j1=-1, int i2=-1, int j2=-1);
  Status medianfilter(unsigned int n);
  Status rebin( const Image &im, unsigned int n );
  Status getradec(double x, double y, double *alpha, double *delta) const;
  Status getxy( double alpha, double delta, double *x, double *y) const;
  Status getPixelsize( double *xscale, double *yscale) const;


  // In Image_fft:
  Status fft( void );
  Status convolve( Image *psf, bool newfft=false);
  Status convolve(const Image &im, Image *psf, bool newfft=false);
  Status makeGaussFT( const Image &im, float sigmax, float sigmay);
  Status mexicanhatFT(const Image &im, float sigmax, float sigmay );
  Status makeRingFT( const Image &im, float sigmax, float sigmay);

  // wavelet specific functions:
  Status maxima( unsigned  int *imax,  unsigned int *jmax, float *val );
  Status mhfilter(const Image &im, float  sigmax, float sigmay,
		  int i1=-1, int j1=-1,int i2=-1,int j2=-1 );
  Status negMhfilter(const Image &im, float  sigmax, float sigmay,
		     int i1=-1, int j1=-1,int i2=-1,int j2=-1 );
  Status centroid(float x0, float y0, float rad, float *xc, float *yc, 
		  float *max, float *xyc, float *x2c, float *y2c ) const;
  Status addgaussian(float x, float y, float norm,
		     float sigmax, float sigmay, float rot);
  Status exposuremask(const Image &expmap);

  string instrument, telescope, detnam, datamode, filter, obs_mode;
  string obs_id, exp_id, content, origin, date, object;
  string pradecsys, observer, obsdate, enddate;
  real32 ra_pnt, dec_pnt, pa_pnt, pequinox, pexposure;

 protected:

 private:

  float *pixel; 
  unsigned int nx, ny; // nim gives size of imaginary part
  char classid[12];
  unsigned int imgbin;

  fftwf_complex *cpixel; // array to store fourier space images
  unsigned long ncomplex;           //  size of complex pixels used by FFTW
  bool ftransformed, fftplanexist;
  // Modification needed? (04/03/09) - JR (AIP) -----//
  //rfftwnd_plan fftplan;
  //fftw_direction fftwdir;
  //  typedef int fft_plan;
  typedef int fftw_direction;
  // changed 7/4/09
  fftwf_plan fftplan;
  fftwf_plan fftplan_r2c;
  fftwf_plan fftplan_c2r;
  fftw_direction fftwdir;
  //-------------------------------------------------//


  int chan_min, chan_max;
  real32 crval1, crval2, crpix1, crpix2, cdelt1, cdelt2, crota2, longpole;
  real64 pc001001, pc001002, pc002002, pc002001;
  real32 drpix1, drpix2, drdelt1, drdelt2, roll_nom;
  real32 blank, bscale, bzero, dec_nom, ra_nom, equinox;
  real32 mjd_obs;
  real64 exposure, ontime, deadtime_corr, start_time, stop_time;
  Projection projectiontype;
  Coordinate_type xy_type;
  string ctype1, ctype2, radecsys;
  
  Status allocate(  unsigned int n,  unsigned int m, float initval = 0.0);
  Status allocate_complex( fftw_direction dir=FFTW_REAL_TO_COMPLEX  );
  Status deallocate();
  
};

//
// Convert from one coordinate system to another
//
// all input value are in radians
inline void ch_coord( double phi, double theta, double phip, double ap, 
		      double dp,  double *alpha, double *delta)
{
  double csina, ccosa;
  //
  phi -= phip;

  *delta = asin( sin(theta) * sin( dp ) + cos(theta) * cos( dp ) * cos( phi) );
  csina =  -cos(theta) * sin( phi );
  ccosa = sin(theta) * cos( dp ) - cos(theta) * sin(dp) * cos( phi);

  // ccosa =cos(*delta) cos(*alpha -ap); 
  // csina= cos(*delta) sin(*alpha -ap)

  *alpha = atan2( csina, ccosa) + ap;

}//ch_coord

#endif
