/*****************************************************************************/
//
// FILE        :  Image_io.cc
// AUTHOR      :  Jacco Vink 
// DESCRIPTION :  
//
//
// COPYRIGHT   :  Jacco Vink, AIP, 1999
//
/*****************************************************************************/
#include "Image.h"


/****************************************************************************/
//
// FUNCTION    :  readImage()
// DESCRIPTION :  Read image from a fits file.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    filename          I
//
//  RETURNS    :  SUCCESS or FAILURE
//
/****************************************************************************/
Status Image::readImage(const string &setname, unsigned int blocknr )
{
  DataSet *imageSet = dataSetServer->open(setname,Dal::Read);
  Array *imageArray = imageSet->array(blocknr);
  unsigned int ndim = imageArray->dimensions();
  vector <unsigned long> size = imageArray->size();

  if( ndim != 2 || size[0] < 2 || size[1] < 2 ){
    dataSetServer->close(imageSet);
    return FAILURE;
  }

  if ( imageSet->hasAttribute( "TELESCOP" ) == true )
    instrument = imageSet->attribute("TELESCOP")->asString();
  if ( imageSet->hasAttribute( "INSTRUME" ) == true )
    instrument = imageSet->attribute("INSTRUME")->asString();
  if ( imageSet->hasAttribute( "DETNAM" ) == true )
    detnam = imageSet->attribute("DETNAM")->asString();
  if ( imageSet->hasAttribute( "DATAMODE" ) == true )
    datamode = imageSet->attribute("DATAMODE")->asString();
  if ( imageSet->hasAttribute( "FILTER" ) == true )
    filter = imageSet->attribute("FILTER")->asString();
  if ( imageSet->hasAttribute( "OBS_MODE" ) == true )
    obs_mode = imageSet->attribute("OBS_MODE")->asString();
  if ( imageSet->hasAttribute( "OBS_ID" ) == true )
    obs_id = imageSet->attribute("OBS_ID")->asString();
  if ( imageSet->hasAttribute( "EXP_ID" ) == true )
    exp_id = imageSet->attribute("EXP_ID")->asString();
  if ( imageSet->hasAttribute( "CONTENT" ) == true )
    content = imageSet->attribute("CONTENT")->asString();
  if ( imageSet->hasAttribute( "ORIGIN" ) == true )
    origin = imageSet->attribute("ORIGIN")->asString();
  if ( imageSet->hasAttribute( "OBJECT" ) == true )
    object = imageSet->attribute("OBJECT")->asString();
  if ( imageSet->hasAttribute( "RA_PNT" ) == true )
    ra_pnt = imageSet->attribute("RA_PNT")->asReal();
  if ( imageSet->hasAttribute( "DEC_PNT" ) == true )
    dec_pnt = imageSet->attribute("DEC_PNT")->asReal();
  if ( imageSet->hasAttribute( "PA_PNT" ) == true )
    pa_pnt = imageSet->attribute("PA_PNT")->asReal();
    if ( imageSet->hasAttribute( "DATE-OBS" ) == true )
    obsdate = imageSet->attribute("DATE-OBS")->asString();
  if ( imageSet->hasAttribute( "DATE-END" ) == true )
    enddate = imageSet->attribute("DATE-END")->asString();
  if ( imageSet->hasAttribute( "OBSERVER" ) == true )
    observer = imageSet->attribute("OBSERVER")->asString();
  if ( imageSet->hasAttribute( "EXPOSURE" ) == true ){
    pexposure = imageSet->attribute("EXPOSURE")->asReal();
    exposure = pexposure;
  }

  if ( imageSet->hasAttribute( "EQUINOX" ) == true ){
    equinox = imageSet->attribute("EQUINOX")->asReal();
    pequinox = imageSet->attribute("EQUINOX")->asReal();
  }
  
  if ( imageSet->hasAttribute( "RADECSYS" ) == true ){
    radecsys = imageSet->attribute("RADECSYS")->asString();
    pradecsys = imageSet->attribute("RADECSYS")->asString();
  }//if

  
  ArrayData *imData = imageArray->data();


  // In the C++ versions of imageArray and addArray 
  // nx corresponds to size[1] and ny to size[0] !!!!!!
   
  allocate( size[1], size[0]);


  switch( imageArray->dataType() ){
  case Dal::Int8:
    {
      int8 *p = imData->int8Data();
      for( unsigned long i=0; i < nx*ny ; i++)
	pixel[i] = (real32)p[i];
    }
    break;
  case Dal::Uint16:
    {
      uint16  *p = imData->uint16Data();
      for( unsigned long i=0; i < nx*ny ; i++)
	pixel[i] = (real32)p[i];
    }
    break;
  case Dal::Int16:
    {
      int16 *p = imData->int16Data();
      for( unsigned long i=0; i < nx*ny ; i++)
	pixel[i] = (real32)p[i];
    }
    break;
  case Dal::Uint32:
    {
      uint32 *p = imData->uint32Data();
      for( unsigned long i=0; i < nx*ny ; i++)
	pixel[i] = (real32)p[i];
    }
    break;
  case Dal::Int32:
    {
      int32 *p = imData->int32Data();
      for( unsigned long i=0; i < nx*ny ; i++)
	pixel[i] = (real32)p[i];
    }
    break;
  case Dal::Real32:
    {
      real32 *p = imData->real32Data();
      for( unsigned long i=0; i < nx*ny ; i++)
	pixel[i] = p[i];
    }
    break;
  case Dal::Real64:
    {
      real64 *p = imData->real64Data();
      for( unsigned long i=0; i < nx*ny ; i++)
	pixel[i] = (real32)p[i];
    }
    Sas::warning("unsupportedDataType","The image array is in real64 and will be casted to real32, this may lead to unwanted results");
    break;
  default:
    return FAILURE;
  }//case

  //  delete imageArray;

  Attributable *attributes;
  if( imageArray->hasAttribute("crpix1") == true ) // follows dal convention
    attributes = imageArray;
  else
    attributes = imageSet;

  if ( attributes->hasAttribute( "CTYPE1" ) == true ){
    ctype1 = attributes->attribute("CTYPE1")->asString();
    ctype2 = attributes->attribute("CTYPE2")->asString();
  }//if

  crpix1 = attributes->attribute("crpix1")->asReal();
  crpix2 = attributes->attribute("crpix2")->asReal();
  cdelt1 = attributes->attribute("cdelt1")->asReal();
  cdelt2 = attributes->attribute("cdelt2")->asReal();
  crval1 = attributes->attribute("crval1")->asReal();
  crval2 = attributes->attribute("crval2")->asReal();
  if ( attributes->hasAttribute( "pc001001" ) == true ){ // rotation matrix
    pc001001 =  attributes->attribute("pc001001")->asReal();
    pc001002 =  attributes->attribute("pc001002")->asReal();
    pc002002 =  attributes->attribute("pc002002")->asReal();
    pc002001 =  attributes->attribute("pc002001")->asReal();
  }
  else{
    if( attributes->hasAttribute( "crota2" ) == true )
      crota2 =  attributes->attribute("crota2")->asReal();
    else
      crota2 = 0.0;
    float s=1.0;
    if( cdelt1 != 0.0 && cdelt2 != 0.0 )
      s = cdelt2/cdelt1;
    pc001001 =  cos(crota2*DEG2RAD);
    pc001002 = -sin(crota2*DEG2RAD) * s;
    pc002001 =  sin(crota2*DEG2RAD) / s;
    pc002002 =  cos(crota2*DEG2RAD);
  }//if
  
  if ( attributes->hasAttribute( "EQUINOX" ) == true ){
    equinox = attributes->attribute("EQUINOX")->asReal();
    pequinox = attributes->attribute("EQUINOX")->asReal();
  }
  
  if ( attributes->hasAttribute( "RADECSYS" ) == true ){
    radecsys = attributes->attribute("RADECSYS")->asString();
    pradecsys = attributes->attribute("RADECSYS")->asString();
  }//if
  
  if (  attributes->hasAttribute( "EXPOSURE" ) == true )
    exposure =  attributes->attribute("EXPOSURE")->asReal();

  delete imData;
  dataSetServer->close(imageSet);

  return SUCCESS;
}//readImage()

/****************************************************************************/
//
// FUNCTION    :  writeImage()
// DESCRIPTION :  Write image to a fits file.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    filename          I
//
//  RETURNS    :  SUCCESS or FAILURE
//
/****************************************************************************/
Status Image::writeImage(const string &setname) const
{
  if( nx < 1 || ny < 1 ){
    //    warning("noData","Image contains no data");
    return FAILURE;
  }//if

  DataSet *imageSet = dataSetServer->open(setname,Dal::Create);
  vector <unsigned long> size(2);
  // In the C++ versions of imageArray and addArray 
  // nx corresponds to size[1] and ny to size[0] !!!!!!
  size[0] = ny;
  size[1] = nx;
  Array *imageArray = imageSet->addArray("PRIMARY",Dal::Real32,size);

  *imageSet->addAttribute("CRPIX1","Reference coordinate pixel")=crpix1;
  *imageSet->addAttribute("CRPIX2","Reference coordinate pixel")=crpix2;
  *imageSet->addAttribute("CRVAL1","Coordinates of ref. pixel","deg")=crval1;
  *imageSet->addAttribute("CRVAL2","Coordinates of ref. pixel","deg")=crval2;
  *imageSet->addAttribute("CDELT1","Pixel size")=cdelt1;
  *imageSet->addAttribute("CDELT2","Pixel size")=cdelt2;
  *imageSet->addAttribute("PC001001","Element of rotation matrix")= pc001001;
  *imageSet->addAttribute("PC001002","Element of rotation matrix")= pc001002;
  *imageSet->addAttribute("PC002001","Element of rotation matrix")= pc002001;
  *imageSet->addAttribute("PC002002","Element of rotation matrix")= pc002002;
  *imageSet->addAttribute("CROTA2","image rotation","deg")=crota2;
  *imageSet->addAttribute("EQUINOX","Pixel size")=equinox;
  *imageSet->addAttribute("CTYPE1","Coordinate type ")= string(ctype1);
  *imageSet->addAttribute("CTYPE2","Coordinate type ")= string(ctype2);
  *imageSet->addAttribute("RADECSYS","Coordinate system")= radecsys;
  *imageSet->addAttribute("TELESCOP","Telescope (mission) name")= telescope;
  *imageSet->addAttribute("INSTRUME","Instrument name")= instrument;
  *imageSet->addAttribute("DETNAM","Prime or redundant")= detnam;
  *imageSet->addAttribute("DATAMODE","Instr. mode (IMAGING,TIMING,..)")= datamode;
  *imageSet->addAttribute("FILTER","Filter ID")= filter;
  *imageSet->addAttribute("OBS_MODE","Obs. mode (pointing or slew)")= obs_mode;
  *imageSet->addAttribute("OBS_ID","Observation identifier")= obs_id;
  *imageSet->addAttribute("EXP_ID","Exposure identifier")= exp_id;
  *imageSet->addAttribute("CONTENT","Content of file")= content;
  //  *imageSet->addAttribute("ORIGIN","")= origin;
  *imageSet->addAttribute("OBJECT","Name of observed object") = object;
  *imageSet->addAttribute("RA_PNT","RA of instr. pointing") = ra_pnt;
  *imageSet->addAttribute("DEC_PNT","DEC of instr. pointing") = dec_pnt;
  *imageSet->addAttribute("PA_PNT","RA of instr. pointing") = ra_pnt;
  *imageSet->addAttribute("EXPOSURE","Exposure (sec)") = exposure;
  *imageSet->addAttribute("DATE-OBS","Start of the observation","D/M/Y") 
    = obsdate;
  *imageSet->addAttribute("DATE-END","End of the observation","D/M/Y") 
    = enddate;
  *imageSet->addAttribute("OBSERVER","Exposure (sec)") = observer;

  ArrayData *imData = imageArray->data();
  real32 *p = imData->real32Data();
  for( unsigned long i=0; i < nx*ny ; i++)
    p[i] = pixel[i];


  //  delete imData;
  //  delete imageArray;
  dataSetServer->close(imageSet);

  return SUCCESS;
}// writeImage()




/****************************************************************************/
//
// FUNCTION    :  printInfo
// DESCRIPTION :  Printimage information to the screen.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    filename          I
//
//  RETURNS    :  void
//
/****************************************************************************/
void Image::printInfo( void ) const
{
  FILE *testStream = stdout;

  fprintf(testStream,"\n");
  fprintf(testStream,"\tnx         = %6d\n",nx);
  fprintf(testStream,"\tny         = %6d\n",ny);
  fprintf(testStream,"\tblank      = %e\n", blank);
  fprintf(testStream,"\tbscale     = %e\n", bscale);
  fprintf(testStream,"\tbzero      = %e\n", bzero);
  fprintf(testStream,"\tcrval1     = %e\n", crval1 );
  fprintf(testStream,"\tcrval2     = %e\n", crval2 );
  fprintf(testStream,"\tcdelt1     = %e\n", cdelt1 );
  fprintf(testStream,"\tcdelt2     = %e\n", cdelt2);
  fprintf(testStream,"\tcrpix1     = %e\n", crpix1 );
  fprintf(testStream,"\tcrpix2     = %e\n", crpix2 );
  fprintf(testStream,"\tcrota2     = %e\n", crota2 );
  fprintf(testStream,"\tpc001001   = %f\n", pc001001 );
  fprintf(testStream,"\tpc001002   = %f\n", pc001002 );
  fprintf(testStream,"\tpc002002   = %f\n", pc002002 );
  fprintf(testStream,"\tpc002001   = %f\n", pc002001 );
  //  fprintf(testStream,"\tctype1     = %s\n", ctype1 );
  // fprintf(testStream,"\tctype2     = %s\n", ctype2 );
  fprintf(testStream,"\tequinox    = %.1f\n", equinox );
  fprintf(testStream,"\tlongpole   = %.1f\n", longpole );
  fprintf(testStream,"\tprojection = %d\n", projectiontype );
  fprintf(testStream,"\timgbin     = %d\n", imgbin );

  //  if( strncmp(ctype1,"RA",2)== 0 && strncmp(ctype2,"DEC",3)== 0){
  //    fprintf(testStream,"\tradecsys   = %s\n", radecsys );
  //  }

  cout <<  "object     = " << object << endl; 
  cout <<  "ra_nom     = " << ra_nom << endl; 
  cout <<  "dec_nom    = " << dec_nom << endl; 
  cout <<  "telescope  = " << telescope << endl;
  cout <<  "instrument = " << instrument << endl;
  cout <<  "exposure   = " << exposure << endl;
  //  fprintf(testStream,"\texposure   = %.1f\n", exposure );
  //  fprintf(testStream,"\tobsdate    = %s\n", obsdate);
  //  fprintf(testStream,"\tmjd obs    = %.2f\n", mjd_obs);
  fprintf(testStream,"\tminimum    = %.5e\n", minimum());
  fprintf(testStream,"\tmaximum    = %.5e\n", maximum());
  fprintf(testStream,"\ttotal      = %.6e\n", total());
  //fprintf(testStream,"\tcreator    = %s\n", creator);

  return;
}//printInfo()
