//
// Module: ewavelet.cc
//
// Authors:  Jacco Vink - AIP, Georg Lamer - AIP
//         
// Description: Detection of extended EPIC sources using a wavelet algorithm.
// 
// Comments: 
// 
// $Id: module.cc,v 1.1 1998/09/03 13:37:31 gvacanti Exp $
#include <stdio.h>
#include "param.h"
#include "errstr.h"
#include "Image.h"
#include "srclist.h"
#include "ewavelet.h"

Status wavelet_analysis( Image *convim, 
			 Image *convmask, Image *convbkg, Image *image, 
			 Image *countrateim, Image *expmask,
        		 const Image &expmap,
			 Image *background, Image *wavelet,
			 float scale, float prob, float edgethreshold,
                         Sourcelist *srclist);
Status calcBackground( Image *background, 
		       const Image &inputimage, const Image &reconstruction, 
		       float scale, int binf);
Status calcBackground( Image *background, 
		       const Image &inputImage, 
		       const Image &reconstruction, 
		       const Image &expmask, 
		       float scale);
double corr_threshold_damiani( double qbgd, double sign );
double corr_threshold( double qbgd, double sign );
Status reconstructImage( const Image &image, const Image &expmap,
			 const Sourcelist &srclist,Image *reconstruction );
Status reconstructImage( const Sourcelist &srclist, float maxscale, 
			 Image *reconstruction);

#define  MINFFTSCALE 1.3 // minimum scale for which to use a FFT
// #define EDGETHRESHOLD 5.0e-2 defines for what wavelet signal for the 
                             // exposure mask one can trust a detection or not


void ewavelet()
{


  int niter;
  //  char message[81];
  string imageset, expmapset, srclistset, inputlist;
  float minscale, maxscale, threshold, edgethreshold, scale;
  Image image, expmap, expmask, countrateim, background;
  Image convim, convmask, convbkg;
  Image wavelet, correctedImage, reconstruction;
  Sourcelist srclist(50000), templist(50000);
  bool withinputlist;  
  
  withinputlist = booleanParameter("withinputlist"); 
  inputlist  = stringParameter("inputlist");
  imageset   = stringParameter("imageset");
  expmapset   = stringParameter("expmapset");
  srclistset = stringParameter("srclistset");
  minscale = realParameter("minscale");
  maxscale = realParameter("maxscale");
  threshold = realParameter("threshold");
  niter = intParameter("niter");
  edgethreshold = realParameter("edgethreshold");


  if( image.readImage( imageset ) == FAILURE ){
    errstr << "Error reading input image " << Sas::error("inputImError");
  }//if


  if(  booleanParameter("useexpmap") == true ){
    if( expmap.readImage( expmapset) == FAILURE )
      errstr << "Error reading exposure map " << Sas::error("inputMapError");
  }
  else{
    expmap = image;
    expmap *= 0.0;
    expmap += image.getExposure();
    Sas::warning("noExpMap","No proper exposure map will be used, this can lead to detection of spurious sources/wrong count rates.");
  }

  if( image.obs_id != expmap.obs_id || image.exp_id != expmap.exp_id || 
      image.instrument != expmap.instrument ){
    Sas::warning("noMatch","The OBS_ID,EXP_ID and/or INSTRUMENT of the exposure map and image do not match");
  }//if

  expmask.exposuremask( expmap );





  if (withinputlist == true){
      errstr << "Creating background and/or reconstruction image from input list: " <<  inputlist << 
      Sas::message(Msg::App, Msg::Verbose); 
   
      if (booleanParameter("makerecon") == false && booleanParameter("savebkgrnd") == false){
      Sas::error("paramError","withinputlist=yes, makerecon=no, savebkgrnd=no -> can't do anything");
      } 

      errstr << "input image   : " << imageset << 
        Sas::message(Msg::App, Msg::Verbose);
      errstr << "input source list   : " << inputlist << 
        Sas::message(Msg::App, Msg::Verbose);
      errstr << "maximum wavelet scale = " << maxscale << 
        Sas::message(Msg::App, Msg::Verbose);


      // read the input source list:
      if ( srclist.readset(inputlist) == FAILURE){
         errstr << "Error reading source list " << Sas::error("inputImError");
      }//if

 
      scale=maxscale;
      reconstruction = image;
      reconstruction *=  0;

      // create the reconstruction image: 
      if( reconstructImage( srclist,1.1*scale, &reconstruction ) 
	  == FAILURE){
          Sas::error("reconError","Error making a reconstructed image" );
      }
      if(  booleanParameter("makerecon") == true ){
          string recimageset = stringParameter("recimageset");
          Sas::message(Msg::App, Msg::Verbose,"Saving reconstructed image.");
          errstr << "Reconstructed image name : " << recimageset << 
            Sas::message(Msg::App, Msg::Verbose);
          reconstruction.writeImage(recimageset);
      }//if
  
      if( booleanParameter("savebkgrnd") == true ){
      // make a background image for this scale
          if(calcBackground( &background, image, reconstruction, expmask,
	        	      scale) == FAILURE ){
	      Sas::error("bgdCalc","Error generating background map.");
          }
          string bkgrndset = stringParameter("bkgrndset");
          Sas::message(Msg::App, Msg::Verbose,"Saving last background map.");
          errstr << "Bkg image name : " << bkgrndset << 
            Sas::message(Msg::App, Msg::Verbose);
          background.writeImage(bkgrndset);
      }//if

      // in mode withinputlist=yes, we are done here
      return;
  }


  errstr << "input image   : " << imageset << 
    Sas::message(Msg::App, Msg::Verbose);
  errstr << "exposure map : " << expmapset << 
    Sas::message(Msg::App, Msg::Verbose);
  errstr << "source list   : " << srclistset<< 
    Sas::message(Msg::App, Msg::Verbose);
  errstr << "minimum wavelet scale = " << minscale << 
    Sas::message(Msg::App, Msg::Verbose);
  errstr << "maximum wavelet scale = " << maxscale << 
    Sas::message(Msg::App, Msg::Verbose);
  errstr << "detection  threshold  = " << threshold << 
    Sas::message(Msg::App, Msg::Verbose);
  errstr << "number of iterations  = " << niter << 
    Sas::message(Msg::App, Msg::Verbose);

  if( minscale > maxscale )
    errstr <<"minscale should be smaller than maxscale" << Sas::error("paramError");

  countrateim = image;
  countrateim /= expmap;
  reconstruction = image;


  // create FFT's of the image, cout-rate image, and exposure map
  if( maxscale >= MINFFTSCALE )
    if( image.fft() == FAILURE || countrateim.fft() == FAILURE ||
	expmask.fft() == FAILURE ){
      //      errstr 
      cout << "Error using FFT " << endl;
    }//if


  srclist.copyKeywords( image);


  for( int i = 1; i <= niter ; i++){

    scale = minscale;
    srclist.init();

    reconstruction *= 0; // initialize reconstructed image (updated each scale)
    while(scale <= maxscale+0.1 ){

      errstr << "   iteration " << i << ", scale = " << scale <<
      Sas::message(Msg::App,Msg::Verbose);
      
 
      // update the reconstruction image for this scale (only<= scale are used)
      if( reconstructImage( templist,1.1*M_SQRT2*scale, &reconstruction ) 
	  == FAILURE){
	Sas::error("reconError","Error making a reconstructed image " );
      }
 

      // make a background image for this scale
      if(calcBackground( &background, image, reconstruction, expmask,
			 scale) == FAILURE ){
	Sas::error("bgdCalc","Error generating background map. (1)");
      }


      // search for candidate sources
      if( wavelet_analysis(&convim, &convmask, &convbkg,
			   &image, &countrateim, &expmask, expmap, 
			   &background, &wavelet, scale,
			   threshold, edgethreshold, &srclist)==FAILURE){
	Sas::error("wavCalc","Wavelet analysis error.");
      }//if

      scale *= M_SQRT2;

    }//while()


    srclist.smallfirst();

    srclist.calcScale();

    if( i != niter )
      templist.copy( srclist);


  }//for


  srclist.calcCoordinates( image );
  //  srclist.statistics();



  if( srclist.writeset(srclistset) == FAILURE )
    Sas::error("srclistError","Error generating final source list.");


  if(  booleanParameter("makerecon") == true ){
    string recimageset = stringParameter("recimageset");
    //if( reconstructImage( image, expmap,srclist, &reconstruction )==FAILURE){
    //      Sas::error("reconError","Error making a reconstructed image" );
    //    }


    Sas::message(Msg::App, Msg::Verbose,"Saving reconstructed image.");
    errstr << "Reconstructed image name : " << recimageset << 
      Sas::message(Msg::App, Msg::Verbose);
    reconstruction.writeImage(recimageset);
  }//if
  


  if( booleanParameter("savebkgrnd") == true ){
    string bkgrndset = stringParameter("bkgrndset");
    Sas::message(Msg::App, Msg::Verbose,"Saving last background map.");
    errstr << "Bkg image name : " << bkgrndset << 
      Sas::message(Msg::App, Msg::Verbose);
    background.writeImage(bkgrndset);
  }//if



  return;
}//ewavelet()

/*****************************************************************************/
//
// FUNCTION    : wavelet_analysis()
// DESCRIPTION : 
//
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :  SUCCESS or FAILURE
//
/*****************************************************************************/
Status wavelet_analysis( Image *convim, 
			 Image *convmask, Image *convbkg, Image *image, 
			 Image *countrateim, Image *expmask,
			 const Image &expmap,
       			 Image *background, Image *wavelet,
	      		 float scale, float prob, float edgethreshold,
                         Sourcelist *srclist)
{

  unsigned int nx, ny;

  //
  // edgethreshold = edgethreshold * 1.0;
  //
  *convmask = *expmask;

  image->getnxny(&nx, &ny);
  if( scale < 1.0 || prob < 0.1 ){
    //    Sas::warning("wavInput","Incorrect wavelet input parameter.");
    puts("Incorrect wavelet input parameter.");
    return FAILURE;
  }//if


  if( scale <= MINFFTSCALE ){
    if( convim->mhfilter( *image, scale, scale) == FAILURE || 
	//	convcntr->mhfilter( *countrateim, scale, scale) == FAILURE ||
	//	convmask->mhfilter( *expmask, scale, scale) == FAILURE ||
        convbkg->mhfilter( *background, scale, scale) == FAILURE
	){
      //    Sas::warning("wavConv","Error in image convolutions.");
      puts("Error in image convolutions.");
      return FAILURE;
    }//if
  }
  else{
    
    *convbkg=*background;
    wavelet->mexicanhatFT( *image, scale, scale);
    if( convim->convolve( *image, wavelet, false ) == FAILURE ||
	//	convcntr->convolve( *countrateim, wavelet, false) == FAILURE ||
	//	convmask->convolve( *expmask, wavelet, false) == FAILURE ||
        convbkg->convolve(wavelet, false) == FAILURE )
      return FAILURE;
  }//if



  unsigned int imax = 2, jmax = 2, ndet;
  float val, qbgd, bgd, corr1, exposure, corr0;
  float max, xc, yc, xx, yy, xy, radius = 2.0*scale;
  float norm = 4.0 *M_PI * scale*scale;    // wavelet normalization factor
  while( convim->maxima( &imax, &jmax, &corr1 )== SUCCESS &&
	 srclist->ndetect < srclist->nmax ){

    val = corr1 * norm;
    bgd =  background->get( imax, jmax);
    qbgd = 0.5 * norm * bgd;
    exposure = expmap.get( imax, jmax);
    corr0    = corr_threshold_damiani( qbgd, prob )/(2.0 * M_PI*SQR(scale));


    /* val*0.25 >=  corr_threshold(qbgd, prob )*/

    //    if( corr1 > corr0  && fabs(convmask->get(imax, jmax)) < edgethreshold &&
    //	expmask->get(imax, jmax) > 1.0e-3 && imax > 1.5*scale && 
    //	jmax > 1.5*scale && imax < nx-1.5*scale && jmax < ny - 1.5*scale){ 


        if( corr1 > corr0  && corr1 > (edgethreshold * fabs(convbkg->get(imax, jmax))) &&
    	expmask->get(imax, jmax) > 1.0e-3 && imax > 1.5*scale && 
    	jmax > 1.5*scale && imax < nx-1.5*scale && jmax < ny - 1.5*scale){

 //        if( corr1 > corr0  &&
 //    	expmask->get(imax, jmax) > 1.0e-3 && imax > 1.5*scale && 
 //    	jmax > 1.5*scale && imax < nx-1.5*scale && jmax < ny - 1.5*scale){

        if( countrateim->centroid(float(imax), float(jmax), radius, 
     			 &max,&xc, &yc, &xy, &xx, &yy) == SUCCESS ){
	srclist->ndetect += 1;
	ndet = srclist->ndetect-1;
	srclist->ipix[ndet]  = imax;
	srclist->jpix[ndet]  = jmax;
	srclist->xima[ndet]  = xc;
	srclist->yima[ndet]  = yc;
	srclist->xerr[ndet]  = sqrt(xx/val);
	srclist->yerr[ndet]  = sqrt(yy/val);
	srclist->corr[ndet]  = corr1;
	//	srclist->crcorr[ndet]= convcntr->get(imax, jmax);
	srclist->scale[ndet] = scale;
	srclist->sigma[ndet] = scale;
	srclist->corr0[ndet] = corr0;
	srclist->background[ndet] = bgd;

	if( xx*yy > 0.0)
	  srclist->xycorr[ndet]= xy/sqrt(xx*yy);
	else
	  srclist->xycorr[ndet]= 0.0;
	srclist->cnts[ndet]    = corr1*norm;
	srclist->cntserr[ndet] = sqrt(val);
	//	srclist->cntrate[ndet] = srclist->crcorr[ndet]*norm;
	srclist->exposure[ndet] = exposure;
	srclist->ignore[ndet]  = false;
        }//if

    }//if
    //
    // update the pixel coordinates for searching for maxima
    if( imax != (unsigned int)nx )
      imax++;
    else{
      imax = 1;
      jmax++;
    }//else


  }//while

  if( srclist->ndetect >= srclist->nmax){
    puts("wavelet_analysis(): Maximum number of sources reached!");
    return FAILURE;
  }//

  return SUCCESS;

}// wavelet_analysis()


/*****************************************************************************/
//
// FUNCTION    : corr_threshold_damini( )
// DESCRIPTION : Correlation threshold after Damiani et al. 1997, ApJ 483, 353
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : threshold
//
/*****************************************************************************/
double corr_threshold_damiani( double qbgd, double ksigma )
{

  const double c1= -0.2336, c2 = 0.0354, c3 = 0.1830;
  double threshold;

  // note that Damiani define the 1 without factor 2pi, but multiply
  // by 2 pi within their formula (7)

  threshold = ksigma * sqrt( qbgd ) + (c1 + c2*ksigma + c3*SQR(ksigma));

  return threshold;

}//corr_threshold_damiani( )




/*****************************************************************************/
//
// FUNCTION    : corr_threshold( )
// DESCRIPTION : 
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : threshold
//
/*****************************************************************************/
double corr_threshold( double qbgd, double sign )
{
  double lsign = log10( sign );
  double alo=0.00462, blo =0.0661;
  double clo= -0.0154*SQR( lsign ) -0.252 * lsign -0.031;
  double ahi, bhi = -1.115 * lsign - 1.038; 
  double lq, lc0, c0;

  if( qbgd > 0.0)
    lq =log10( qbgd );
  else
    lq = -1.0e30;
  if( lq > 0.0 ){
    ahi = -0.509 * lsign + 1.897;
    if( lsign > -7.0 )
      ahi +=   - 0.00172* pow( lsign + 7.0, 3.606 );

    c0 = ahi * sqrt(qbgd) + bhi;
  }else{
    lc0 = alo * lq*lq + blo * lq + clo;
    c0 = exp( lc0*log(10.0));
  }//if


  return c0;
}//corr_threshold()
